/**
 * @copyright Copyright 2013 by the Websoft Research Group, Nanjing University, PR China
 * @license MIT (https://opensource.org/licenses/MIT)
 * @link http://ws.nju.edu.cn/ptime
 */
package cn.edu.nju.ws.ptime;

import cn.edu.nju.ws.ptime.core.ExpressionExtractor;
import cn.edu.nju.ws.ptime.core.PatternManager;
import cn.edu.nju.ws.ptime.tml.ExpressionIndex;
import cn.edu.nju.ws.ptime.tml.TDocument;
import cn.edu.nju.ws.ptime.tml.TMLParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.Tuple2;

import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;


public class Controller {
	private static final Logger LOGGER = LoggerFactory.getLogger(Controller.class);

	/**
	 * For quickly determining the value of \rho. Read *.tml files from $dataDir/$fold/train/ and $dataDir/$fold/dev/.
	 * print annotated files to $outputDir/$fold/ ($fold <- (0, $folds), $rho <- ($rhoMin / 100, $rhoMax / 100, $rhoStep / 100) ).
	 *
	 * @param dataDir where to read training and developing files.
	 * @param outputDir where to output the running results
	 * @param folds number of folds
	 * @param rhoMin range of \rho (x100)
	 * @param rhoMax range of \rho (x100)
	 * @param rhoStep range of \rho (x100)
	 * @throws IOException
	 */
	public static void crossValidation(String dataDir, String outputDir, int folds, int rhoMin, int rhoMax, int rhoStep) throws
			IOException {
		for (int fold = 0; fold < folds; ++fold) {
			String trainingPath = String.format("%s/%d/train", dataDir, fold);
			LOGGER.info("Reading training documents from {} ...", trainingPath);
			File trainingDir = new File(trainingPath);
			File[] trainingFiles = trainingDir.listFiles(s -> s.getName().endsWith(".tml"));
			assert trainingFiles != null;
			TDocument[] trainDocs = new TDocument[trainingFiles.length];
			ArrayList<ExpressionIndex> texBuffer = new ArrayList<>();
			for (int i = 0; i < trainDocs.length; ++i) {
				Tuple2<TDocument, ExpressionIndex[]> parseResult = TMLParser.parseTML(trainingFiles[i], i);
				trainDocs[i] = parseResult._1;
				texBuffer.addAll(Arrays.asList(parseResult._2));
			}
			ExpressionIndex[] trainTexs = texBuffer.toArray(new ExpressionIndex[0]);

			LOGGER.info("Generating candidate patterns ...");
			String[][] candidatePatterns = PatternManager.generatePatterns(trainDocs, trainTexs);

			LOGGER.info("Assessing patterns on training set ...");
			Map<Integer, String[][]>
					patternRecords = PatternManager.selectPatterns(rhoMin, rhoMax, rhoStep, trainDocs, candidatePatterns, trainTexs);


			String devPath = String.format("%s/%d/dev", dataDir, fold);
			LOGGER.info("Reading test documents from {} ...", devPath);
			File testDir = new File(devPath);
			File[] testFiles = testDir.listFiles(s -> s.getName().endsWith(".tml"));
			assert testFiles != null;
			TDocument[] testDocs = new TDocument[testFiles.length];
			for (int i = 0; i < testDocs.length; ++i) {
				Tuple2<TDocument, ExpressionIndex[]> parseResult = TMLParser.parseTML(testFiles[i], i);
				testDocs[i] = parseResult._1;
			}

			LOGGER.info("Extracting time expressions from test documents ...");
			for (int x = rhoMin; x <= rhoMax; x += rhoStep) {
				File output = new File(String.format("%s/%d/%.2f/", outputDir, fold, x/100.0));
				if (!output.exists()) output.mkdir();
				apply(testDocs, output, patternRecords.get(x));
			}
		}
	}

	private static void apply(TDocument[] tDocuments, File outputDir, String[][] patterns) throws IOException {
		ExpressionExtractor expressionExtractor = new ExpressionExtractor(patterns);
		for (TDocument tDoc: tDocuments) {
			List<ExpressionIndex> texes = expressionExtractor.extractForRecognition(tDoc);
			BufferedWriter writer = new BufferedWriter(new FileWriter(String.format("%s/%s", outputDir.getPath(), tDoc.docName())));
			writer.write(TMLParser.transformAnnotations(tDoc, texes));
			writer.close();
		}
	}

	/**
	 * Use generated patterns to extract time expression from test documents.
	 *
	 * @param testDir where to read the *.tml files
	 * @param patternFile where to read the patterns
	 * @param outputDir where to output the annotated results
	 * @throws IOException
	 */
	public static void test(File testDir, File patternFile, File outputDir) throws IOException {
		LOGGER.info("Reading test documents from {} ...", testDir.getPath());
		File[] testFiles = testDir.listFiles(s -> s.getName().endsWith(".tml"));
		assert testFiles != null;
		TDocument[] testDocs = new TDocument[testFiles.length];
		for (int i = 0; i < testDocs.length; ++i) {
			Tuple2<TDocument, ExpressionIndex[]> parseResult = TMLParser.parseTML(testFiles[i], i);
			testDocs[i] = parseResult._1;
		}

		LOGGER.info("Reading patterns from {} ... ", patternFile.getName());
		BufferedReader reader = new BufferedReader(new FileReader(patternFile));
		String[][] patterns = reader.lines().map(l -> l.split("\t")).toArray(String[][]::new);

		LOGGER.info("Extracting time expressions from test documents ...");
		apply(testDocs, outputDir, patterns);
	}

	/**
	 * Read *.tml documents, generating and selecting patterns.
	 *
	 * @param trainingDir where to read the training *.tml files
	 * @param outputDir where to output the patterns
	 * @param rho the controlling parameter
	 * @throws IOException
	 */
	public static void train(File trainingDir, File outputDir, double rho) throws
			IOException {
		LOGGER.info("Reading training documents from {} ...", trainingDir.getPath());
		File[] trainingFiles = trainingDir.listFiles(s -> s.getName().endsWith(".tml"));
		assert trainingFiles != null;
		TDocument[] trainDocs = new TDocument[trainingFiles.length];
		ArrayList<ExpressionIndex> texBuffer = new ArrayList<>();
		for (int i = 0; i < trainDocs.length; ++i) {
			Tuple2<TDocument, ExpressionIndex[]> parseResult = TMLParser.parseTML(trainingFiles[i], i);
			trainDocs[i] = parseResult._1;
			texBuffer.addAll(Arrays.asList(parseResult._2));
		}

		LOGGER.info("Generating candidate patterns ...");
		ExpressionIndex[] trainTexs = texBuffer.toArray(new ExpressionIndex[0]);
		String[][] candidatePatterns = PatternManager.generatePatterns(trainDocs, trainTexs);

		LOGGER.info("Assessing patterns on training set ...");
		String[][] patterns = PatternManager.selectPatterns(rho, trainDocs, candidatePatterns, trainTexs);

		LOGGER.info("Save selected patterns in {} ...", outputDir);
		BufferedWriter writer1 = new BufferedWriter(new FileWriter(String.format("%s/patterns.txt", outputDir.getPath())));
		BufferedWriter writer2 = new BufferedWriter(new FileWriter(String.format("%s/patterns.readable.txt", outputDir.getPath())));
		for (String[] pattern : patterns) {
			writer1.write(String.join("\t", Arrays.asList(pattern)));
			writer2.write(PatternManager.toReadableString(pattern));
			writer1.newLine();
			writer2.newLine();
		}
		writer1.close();
		writer2.close();
	}

	public static void main(String[] args) throws IOException {
/*
		//For quickly generate cross validation result to determine the value of \rho
		crossValidation("resources/CV-10/TimeBank","output/CV-10/TE", 10, 70, 100, 1);
		crossValidation("resources/CV-10/WikiWars","output/CV-10/WW", 10, 70, 100, 1);
		crossValidation("resources/CV-10/Tweets","output/CV-10/TW", 10, 70, 100, 1);
*/

		//TimeEval-3
		train(new File("resources/tempeval3/TimeBank"), new File("output/TE"), 0.88);
		test(new File("resources/tempeval3/te3-platinum"), new File("output/TE/patterns.txt"), new File("output/TE/tml"));

		//WikiWars
		train(new File("resources/wikiwars/trainingset"), new File("output/WW"), 0.94);
		test(new File("resources/wikiwars/testset"), new File("output/WW/patterns.txt"), new File("output/WW/tml"));

		//Tweets
		train(new File("resources/tweets/trainingset"), new File("output/TW"), 0.95 );
		test(new File("resources/tweets/testset"), new File("output/TW/patterns.txt"), new File("output/TW/tml"));
	}
}
