/**
  * @copyright Copyright 2013 by the Websoft Research Group, Nanjing University, PR China
  * @license MIT (https://opensource.org/licenses/MIT)
  * @link http://ws.nju.edu.cn/ptime
  */
package cn.edu.nju.ws.ptime.core

import java.{lang, util}

import cn.edu.nju.ws.ptime.tml.{ExpressionIndex, TDocument, TaggedToken}
import org.slf4j.LoggerFactory

import scala.collection.JavaConverters._

object PatternManager {
  private val LOGGER = LoggerFactory.getLogger(PatternManager.getClass)

  def exp2pat(exp: Array[TaggedToken]): Array[String] = {
    exp.map(tok => {
      val tType = TokenTypes.name2regex.values.find(tType => tok.tok2string matches tType)
      tType.getOrElse(s"${escapeString(tok.lemma)}/[^\\s]+")
    })
  }

  def generatePatterns(tDocs: Array[TDocument], goldTexs: Array[ExpressionIndex]): Array[Array[String]] = {
    val texContents = goldTexs.map(index => tDocs(index.docId).getExpression(index.from, index.until))
    LOGGER.info("Abstract token to types ...")
    texContents
      .map(tex => exp2pat(tex).mkString("@@@"))   //a dirty hack?
      .distinct
      .map(pStr => pStr.split("@@@"))
  }

  val escape = Array("\\", "$", "(", ")", "[", "]", "{", "}", "*", "+", ".", "?", "^", "|")

  private def escapeString(rawStr: String) = {
    escape.foldLeft(rawStr)((str, chr) => {
      str.replace(chr, s"\\$chr")
    })
  }

  def toReadableString(pattern: Array[String]): String = {
    pattern.map(regex => {
      val defaultName = if (regex.indexOf('/') < 0) regex else regex.substring(0, regex.indexOf('/'))
      TokenTypes.regex2name.getOrElse(regex, defaultName)
    }).mkString(" ")
  }

  def extractExpressions(tDocs: Array[TDocument], patterns: Array[Array[String]]): Array[Array[ExpressionIndex]] = {
    val extractor = new ExpressionExtractor(patterns)
    LOGGER.info("Extract expressions by patterns ...")
    extractor.extractForMeasurement(tDocs).asScala.map(_.asScala.toArray.sorted).toArray
  }

  def selectPatterns(rhoMin: Int, rhoMax: Int, rhoStep: Int, tDocs: Array[TDocument], candidates: Array[Array[String]], goldTexs: Array[ExpressionIndex]): util.Map[lang.Integer, Array[Array[String]]] = {
    val extracted: Array[Array[ExpressionIndex]] = extractExpressions(tDocs, candidates)
    val selector = new PatternSelector(extracted, goldTexs)
    selector.selectGrid(rhoMin, rhoMax, rhoStep, new util.HashSet[Integer]()).asScala.map(result => {
      (result._1, result._2.asScala.map(candidates(_)).toArray)
    }).toMap.asJava
  }

  def selectPatterns(rho: Double, tDocs: Array[TDocument], candidates: Array[Array[String]], goldTexs: Array[ExpressionIndex]): Array[Array[String]] = {
    val extracted: Array[Array[ExpressionIndex]] = extractExpressions(tDocs, candidates)
    val selector = new PatternSelector(extracted, goldTexs)
    selector.select(rho, new util.HashSet[Integer]()).asScala.map(candidates(_)).toArray
  }
}
