/**
 * @copyright Copyright 2013 by the Websoft Research Group, Nanjing University, PR China
 * @license MIT (https://opensource.org/licenses/MIT)
 * @link http://ws.nju.edu.cn/ptime
 */
package cn.edu.nju.ws.ptime.core;

import cn.edu.nju.ws.ptime.tml.ExpressionIndex;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.Tuple2;
import scala.Tuple3;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

public class PatternSelector {
	private static final Logger LOGGER = LoggerFactory.getLogger(PatternSelector.class);

	private double []cost;
	private double [][]weight;
	private int pSize;
	private int eSize;

	public PatternSelector(ExpressionIndex[][] exTs, ExpressionIndex[] gTs) {
		pSize = exTs.length;
		eSize = gTs.length;
		cost = new double[pSize];
		weight = new double[pSize][eSize];
		for (int pId = 0; pId < pSize; pId++) {
			cost[pId] = 1e-8;
			for (int gL = 0, gR = 1, exId = 0; exId < exTs[pId].length; exId++) {
				ExpressionIndex exTex = exTs[pId][exId];
				while (gL < eSize && gTs[gL].before(exTex)) gL++;
				while (gR < eSize && gTs[gR].beforeOrOverlap(exTex)) gR++;
				boolean coverFlag = false;
				for (int gId = gL; gId < gR; gId++) {
					double contain = (exTex.containedCoverage(gTs[gId]));
					if (contain > 0.0) {
						coverFlag = true;
						weight[pId][gId] = Math.max(weight[pId][gId], contain);
					}
				}
				if (!coverFlag) cost[pId]++;
			}
		}
		LOGGER.info("Set up EBMC for selection. |P|={}, |E|={}", pSize, eSize);
	}

	public List<Integer> select(double rho, Set<Integer> builtIn) {
		List<Tuple3<Integer, Double, Double>> rResult = greedy(rho, builtIn);

		int top = IntStream.range(0, pSize).filter(i -> !builtIn.contains(i)).mapToObj(x ->
				new Tuple2<>(x, Arrays.stream(weight[x]).reduce(Double::sum).getAsDouble())
		).reduce((x, y) -> x._2() > y._2() ? x : y).get()._1();

		if (cost[top] > getBudget(rho)) {
			return rResult.stream().map(Tuple3::_1).collect(Collectors.toList());
		}

		builtIn.add(top);
		List<Tuple3<Integer, Double, Double>> wResult = greedy(rho, builtIn);

		List<Integer> rSelect = rResult.stream().map(Tuple3::_1).collect(Collectors.toList());
		List<Integer> wSelect = wResult.stream().map(Tuple3::_1).collect(Collectors.toList());

		double rGain = rResult.get(rResult.size() - 1)._2();
		double wGain = wResult.get(wResult.size() - 1)._2();

		return rGain > wGain ? rSelect : wSelect;
	}

	public Map<Integer, List<Integer>> selectGrid(int min, int max, int step, Set<Integer> builtIn) {
		Map<Integer, List<Integer>> resultMap = new HashMap<>();

		List<Tuple3<Integer, Double, Double>> rResult = greedy(min / 100.0, builtIn);
		int top = IntStream.range(0, pSize).filter(i -> !builtIn.contains(i)).mapToObj(x ->
			new Tuple2<>(x, Arrays.stream(weight[x]).reduce(Double::sum).getAsDouble())
		).reduce((x, y) -> x._2() > y._2() ? x : y).get()._1();
		builtIn.add(top);
		List<Tuple3<Integer, Double, Double>> wResult = greedy(min / 100.0, builtIn);

		LOGGER.info("Comparing solutions ...");
		int rId = 0, wId = 0;
		List<Integer> rSelect = new ArrayList<>(), wSelect = new ArrayList<>();
		for (int rho = max; rho >= min; rho -= step) {
			double budget = getBudget(rho / 100.0);
			double rGain = -1.0, wGain = -1.0;
			while (rId < rResult.size() && rResult.get(rId)._3() <= budget) {
				rSelect.add(rResult.get(rId)._1());
				rGain = rResult.get(rId)._2();
				rId += 1;
			}
			while (wId < wResult.size() && wResult.get(wId)._3() <= budget) {
				wSelect.add(wResult.get(wId)._1());
				wGain = rResult.get(wId)._2();
				wId += 1;
			}
			resultMap.put(rho, new ArrayList<>(rGain > wGain ? rSelect : wSelect));
		}
		return resultMap;
	}

	/**
	 *
	 * @param rho budget = |E|*(1-rho)
	 * @param builtIn Pre-selected patterns
	 * @return List<(Last selected, current Gain, current Cost)>
	 */
	public List<Tuple3<Integer, Double, Double>> greedy(double rho, Set<Integer> builtIn) {
		LOGGER.info("Selecting patterns ...");
		Set<Integer> candidate = new HashSet<>();
		List<Tuple3<Integer, Double, Double>> selected = new ArrayList<>();

		for (int i = 0; i < pSize; ++i) {
			candidate.add(i);
		}

		double budget = getBudget(rho);
		double totalGain = 0.0, totalCost = 0.0;
		double[] gain = new double[eSize];

		List<Integer> uncovered = new LinkedList<>();
		for (int j = 0; j < eSize; ++j) {
			uncovered.add(j);
		}

		for (Integer pid: builtIn) {
			candidate.remove(pid);
			if (cost[pid] + totalCost > budget) {
				continue;
			}
			totalCost += cost[pid];
			totalGain = updateGain(totalGain, gain, uncovered, pid);
			selected.add(new Tuple3<>(pid, totalGain, totalCost));
		}

		while (!candidate.isEmpty() && !uncovered.isEmpty()) {
			LOGGER.debug("  {} Remaining patterns ...", candidate.size());
			int top = candidate.stream().reduce((x, y) -> {
				double deltaGx = uncovered.stream().map(eid -> Math.max(weight[x][eid] - gain[eid], 0)).reduce(Double::sum).get();
				double deltaGy = uncovered.stream().map(eid -> Math.max(weight[y][eid] - gain[eid], 0)).reduce(Double::sum).get();
				return deltaGx * cost[y] > deltaGy * cost[x] ? x : y;
			}).get();
			candidate.remove(top);
			if (cost[top] + totalCost > budget) {
				continue;
			}
			totalCost += cost[top];
			totalGain = updateGain(totalGain, gain, uncovered, top);
			selected.add(new Tuple3<>(top, totalGain, totalCost));

			LOGGER.debug("   Select p_{}, total Gain={}, total Cost={}", top, totalGain, totalCost);
		}
		return selected;
	}

	private double updateGain(double totalGain, double[] gain, List<Integer> uncovered, Integer pid) {
		for (ListIterator<Integer> it = uncovered.listIterator(); it.hasNext(); ) {
			int eid = it.next();
			if (weight[pid][eid] > gain[eid]) {
				totalGain += weight[pid][eid] - gain[eid];
				gain[eid] = weight[pid][eid];
				if (gain[eid] == 1.0) {
					it.remove();
				}
			}
		}
		return totalGain;
	}

	private double getBudget(double rho) {
		return eSize * (1 - rho) + pSize * 1e-4;
	}
}
