/**
  * @copyright Copyright 2013 by the Websoft Research Group, Nanjing University, PR China
  * @license MIT (https://opensource.org/licenses/MIT)
  * @link http://ws.nju.edu.cn/ptime
  */
package cn.edu.nju.ws.ptime.core

import java.util.regex.Pattern

import cn.edu.nju.ws.ptime.tml.{ExpressionIndex, TaggedToken}

import scala.collection.mutable

/**
  * Created by wtding on 2018/12/18.
  */
object Postprocessor {
  val ORDINAL_REGEX = "([2-9]?1st|[2-9]?2nd|[2-9]?3rd|[4-9]th|1[0-9]th|[2-9]0th|[2-9][4-9]th|100+th|zeroth|first|second|third|fourth|fifth|sixth|seventh|eighth|ninth|tenth|eleventh|twelfth|twentieth|thirteenth|fourteenth|fifteenth|sixteenth|seventeenth|eighteenth|nineteenth|hundredth|thousandth|(twenty|thirty|forty|fifty|sixty|seventy|eighty|ninety)-(first|second|third|fourth|fifth|sixth|seventh|eighth|ninth))"
  val NUM_REGEX = "(\\d+|(twenty|thirty|forty|fifty|sixty|seventy|eighty|ninety)-(one|two|three|four|five|six|seven|eight|nine)|one|two|three|four|five|six|seven|eight|nine|ten|eleven|twelve|thirteen|fourteen|fifteen|sixteen|seventeen|eighteen|nineteen|twenty|thirty|forty|fifty|sixty|seventy|eighty|ninety)"
  val D_START_REGEX = "([0-9]|one|two|three|four|five|six|seven|eight|nine|ten|eleven|twelve|thirteen|fourteen|fifteen|sixteen|seventeen|eighteen|nineteen|twenty|thirty|forty|fifty|sixty|seventy|eighty|ninety)"

  /**
    * Merge overlap/consecutive expressions
    *
    * @param texes not-processed time expressions
    * @return merged and sorted time expressions
    */
  def merge(texes: Seq[ExpressionIndex]) = {
    texes.sorted.foldLeft(List.empty[ExpressionIndex])((buffer, ei) => {
      if (buffer.isEmpty)
        List(ei)
      else if (buffer.head.overlapOrConsecutive(ei))
        ExpressionIndex(0, Math.min(buffer.head.from, ei.from), Math.max(buffer.head.until, ei.until), "", "") :: buffer.tail
      else
        ei :: buffer
    })
  }

  /**
    * Annotates time expressions and post-processing
    *
    * @param rawText  Not marked text
    * @param taggedTokens token sequence of the text
    * @param texes time expressions
    * @return
    */
  def annotateText(rawText: String, taggedTokens: IndexedSeq[TaggedToken], texes: Seq[ExpressionIndex]): String = {
    val stringBuilder = new mutable.StringBuilder(rawText)

    merge(texes).zipWithIndex.foreach(iEi => {
      val begin = taggedTokens(iEi._1.from).charBegin
      val end = taggedTokens(iEi._1.until - 1).charEnd
      stringBuilder.insert(end, "</TIMEX3>")
      stringBuilder.insert(begin, "<TIMEX3 tid=\"t" + (iEi._2 + 1) + "\" type=\"DATE\" value=\"\">")
    })
    var texCnt = texes.size + 1
    var text = stringBuilder.toString().replaceAll(" / ", "/").replaceAll(" (<[^>]+>)?-(<[^>]+>)? ", "$1-$2")

    var m = Pattern.compile(s"$NUM_REGEX( ?- ?| and | or | to )<TIMEX3[^>]+>$D_START_REGEX[^<]*</TIMEX3>").matcher(text)
    while (m.find) {
      text = text.replaceFirst(s"$NUM_REGEX( ?- ?| and | or | to )(<TIMEX3[^>]+type="+"\"([A-Z]+)\""+s"[^>]+>$D_START_REGEX[^<]*</TIMEX3>)", "<TIMEX3 tid=\"t" + texCnt + "\" type=\"$6\" value=\"na\">$1</TIMEX3>$4$5")
      texCnt += 1
    }
    m = Pattern.compile(s"<TIMEX3[^>]+>[^<]*$D_START_REGEX</TIMEX3>( ?- ?| and | or | to )$NUM_REGEX").matcher(text)
    while (m.find) {
      text = text.replaceFirst("(<TIMEX3[^>]+type="+"\"([A-Z]+)\""+s"[^>]+>[^<]*$D_START_REGEX</TIMEX3>)( ?- ?| and | or | to )$NUM_REGEX", "$1$4<TIMEX3 tid=\"t" + texCnt + "\" type=\"$2\" value=\"\">$5</TIMEX3>")
      texCnt += 1
    }
    m = Pattern.compile(s"$ORDINAL_REGEX( ?- ?| and | or | to )<TIMEX3[^>]+>$ORDINAL_REGEX[^<]*</TIMEX3>").matcher(text)
    while (m.find) {
      text = text.replaceFirst(s"$ORDINAL_REGEX( ?- ?| and | or | to )(<TIMEX3[^>]+type="+"\"([A-Z]+)\""+s"[^>]+>$ORDINAL_REGEX[^<]*</TIMEX3>)", "<TIMEX3 tid=\"t" + texCnt + "\" type=\"$6\" value=\"\">$1</TIMEX3>$4$5")
      texCnt += 1
    }

    text
  }
}
