/**
  * @copyright Copyright 2013 by the Websoft Research Group, Nanjing University, PR China
  * @license MIT (https://opensource.org/licenses/MIT)
  * @link http://ws.nju.edu.cn/ptime
  */
package cn.edu.nju.ws.ptime.tml

import java.io.File

import cn.edu.nju.ws.ptime.core.{Postprocessor, Preprocessor}
import edu.stanford.nlp.simple.Document

import scala.collection.mutable
import scala.collection.mutable.ArrayBuffer
import scala.xml.XML
import scala.collection.JavaConverters._

object TMLParser {
  val XML_HEAD = "<?xml version=\"1.0\" ?>\n"

  def transformAnnotations(tDoc: TDocument, texes: java.util.List[ExpressionIndex]): String = {
    val xml = XML.loadString(tDoc.rawContent)
    val textRoot = (xml \ "TEXT").head
    val oldText = textRoot.text
      .replaceAll("((?<![tT]wenty|[tT]hirty|[fF]orty|[fF]ifty|[sS]ixty|[sS]eventy|[eE]ighty|[nN]inety)-)|(-(?!one|two|three|four|five|six|seven|eight|nine|first|second|third|fourth|fifth|sixth|seventh|eighth|ninth))", " - ")
      .replaceAll("(?<![<\\d])/", " / ")
    val newText = Postprocessor.annotateText(oldText, tDoc.taggedTokens, texes.asScala)
    val textStart = tDoc.rawContent.indexOf("<TEXT>")
    val textEnd = tDoc.rawContent.indexOf("</TEXT>")
    s"${tDoc.rawContent.substring(0, textStart + 6)}$newText${tDoc.rawContent.substring(textEnd)}"
      .replaceAll("<MAKEINSTANCE[^>]+/>", "")
      .replaceAll("<TLINK[^>]+/>", "")
      .replaceAll("<SLINK[^>]+/>", "")
      .replaceAll("<ALINK[^>]+/>", "")
  }

  def parseTML(file: File, docID: Int): (TDocument, Array[ExpressionIndex]) = {
    val xml = XML.loadFile(file)

    //parsing dct
    val dctValue = (xml \ "DCT" \ "TIMEX3").head \@ "value"
    val dctContent = (xml \ "DCT" \ "TIMEX3").head.text

    //parsing text
    val textRoot = (xml \ "TEXT").head
    val teRecords = ArrayBuffer.empty[ExpressionIndex]
    val taggedTokens = ArrayBuffer.empty[TaggedToken]
    val stringBuilder = new mutable.StringBuilder()
    for (node <- textRoot.child) {
      val text = node.text
        .replaceAll("((?<![tT]wenty|[tT]hirty|[fF]orty|[fF]ifty|[sS]ixty|[sS]eventy|[eE]ighty|[nN]inety)-)|(-(?!one|two|three|four|five|six|seven|eight|nine|first|second|third|fourth|fifth|sixth|seventh|eighth|ninth))", " - ")
        .replaceAll("(?<![<\\d])/", " / ")
      node match {
        case <TIMEX3>{_}</TIMEX3> =>
          teRecords += ExpressionIndex(0, stringBuilder.length, stringBuilder.length + text.length, node \@ "value", node \@ "type")
        case _ =>
      }
      val tokens = text2token(text, stringBuilder.length)
      taggedTokens ++= tokens
      stringBuilder ++= text
    }
    val processedTokens = Preprocessor.preprocess(taggedTokens.asJava).asScala.toArray
    var l = 0
    var r = 0
    val texs = teRecords.map(tmpTex => {
      while (l < processedTokens.length && processedTokens(l).charBegin < tmpTex.from) l += 1
      while (r < processedTokens.length && processedTokens(r).charEnd <= tmpTex.until) r += 1
      ExpressionIndex(docID, l, r, tmpTex.value, tmpTex.eType)
    }).toArray
    (TDocument(docID, dctValue, dctContent, file.getName, XML_HEAD+xml.toString(), processedTokens), texs)
  }

  private def text2token(text: String, charOffset: Int) = {
    new Document(text).sentences().asScala.flatMap(sent => {
      sent.asCoreLabels(_.words(), _.lemmas(), _.posTags(), _.characterOffsetBegin(), _.characterOffsetEnd())
        .asScala.zipWithIndex.map(
        tok => TaggedToken(tok._1.originalText(), tok._1.lemma(), tok._1.tag(), tok._1.beginPosition() + charOffset, tok._1.endPosition() + charOffset)
      )
    })
  }

  def main(args: Array[String]): Unit = {
    val (x, y) = parseTML(new File("resources/te3-platinum/tml/AP_20130322.tml"), 0)
    val xml = XML.loadString(x.rawContent)
    val text = (xml \ "TEXT").head.text
      .replaceAll("((?<![tT]wenty|[tT]hirty|[fF]orty|[fF]ifty|[sS]ixty|[sS]eventy|[eE]ighty|[nN]inety)-)|(-(?!one|two|three|four|five|six|seven|eight|nine|first|second|third|fourth|fifth|sixth|seventh|eighth|ninth))", " - ")
      .replaceAll("(?<![<\\d])/", " / ")
    println(text.length)
    val tok = x.taggedTokens(26)
    println(tok.charBegin + " " + tok.charEnd)
    println(tok.text + " " + tok.lemma)
    println(text.substring(tok.charBegin, tok.charEnd))
  }
}
